package org.xbib.netty.http.common;

import io.netty.handler.codec.http.HttpHeaderValues;
import java.util.Map;
import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.common.Pair;
import org.xbib.datastructures.common.TreeMultiMap;
import org.xbib.net.PercentDecoder;
import org.xbib.net.PercentEncoder;
import org.xbib.net.PercentEncoders;

import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.MalformedInputException;
import java.nio.charset.StandardCharsets;
import java.nio.charset.UnmappableCharacterException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * A multi-map of HTTP request parameters. Each key references a
 * set of parameters collected from the request during message
 * signing.
 * Every key/value pair will be percent-encoded upon insertion.
 */
@SuppressWarnings("serial")
public class HttpParameters {

    private static final Logger logger = Logger.getLogger(HttpParameters.class.getName());

    private static final char EQUAL_CHAR = '=';

    private static final char AMPERSAND_CHAR = '&';

    private final MultiMap<String, String> multiMap;

    private final PercentDecoder percentDecoder;

    private final PercentEncoder percentEncoder;

    private final CharSequence contentType;

    private final Charset encoding;

    public HttpParameters() {
        this(HttpHeaderValues.APPLICATION_X_WWW_FORM_URLENCODED, StandardCharsets.UTF_8);
    }

    public HttpParameters(CharSequence contentType) {
        this(contentType, StandardCharsets.UTF_8);
    }

    public HttpParameters(CharSequence contentType, Charset charset) {
        this.multiMap = new TreeMultiMap<>();
        this.percentDecoder = new PercentDecoder();
        this.percentEncoder = PercentEncoders.getQueryEncoder(charset);
        this.contentType = contentType;
        this.encoding = charset;
    }

    public MultiMap<String, String> asMultiMap() {
        return multiMap;
    }

    public Map<String, Collection<String>> asMap() {
        return multiMap.asMap();
    }

    public CharSequence getContentType() {
        return contentType;
    }

    public Charset getEncoding() {
        return encoding;
    }

    public String get(String key) {
        Collection<String> collection = multiMap.get(key);
        if (collection != null) {
            Iterator<String> iterator = collection.iterator();
            if (iterator.hasNext()) {
                return iterator.next();
            }
        }
        return null;
    }

    public Collection<String> getAll(String key) {
        return multiMap.get(key);
    }

    public void put(String key, Collection<String> values, boolean percentEncode) {
        multiMap.remove(key);
        for (String v : values) {
            add(key, v, percentEncode);
        }
    }

    /**
     * Convenience method to add a single value for the parameter specified by 'key'.
     *
     * @param key the parameter name
     * @param value the parameter value
     * @return the value
     */
    public HttpParameters addRaw(String key, String value) {
        return add(key, value, false);
    }

    public HttpParameters add(String key, String value) {
        return add(key, value, true);
    }

    /**
     * Convenience method to allow for storing null values. {@link #put} doesn't
     * allow null values, because that would be ambiguous.
     *
     * @param key the parameter name
     * @param nullString can be anything, but probably... null?
     * @return null
     */
    public HttpParameters addNull(String key, String nullString) {
        return addRaw(key, nullString);
    }

    /**
     * Convenience method to add a single value for the parameter specified by
     * 'key'.
     *
     * @param key the parameter name
     * @param value the parameter value
     * @param percentEncode whether key and value should be percent encoded before being
     *        inserted into the map
     * @return the value
     */
    private HttpParameters add(String key, String value, boolean percentEncode) {
        try {
            String k = percentEncode ? percentEncoder.encode(key) : key;
            String v = percentEncode ? percentEncoder.encode(value) : value;
            multiMap.put(k, v);
        } catch (CharacterCodingException e) {
            throw new IllegalArgumentException(e);
        }
        return this;
    }

    public String getAsQueryString(boolean percentEncode) throws MalformedInputException, UnmappableCharacterException {
        List<String> list = new ArrayList<>();
        for (String key : multiMap.keySet()) {
            list.add(getAsQueryString(key, percentEncode));
        }
        return String.join(String.valueOf(AMPERSAND_CHAR), list);
    }

    /**
     * Concatenates all values for the given key to a list of key/value pairs
     * suitable for use in a URL query string.
     *
     * @param key the parameter name
     * @return the query string
     * @throws MalformedInputException if input is malformed
     * @throws UnmappableCharacterException if characters are unmappable
     */
    public String getAsQueryString(String key)
            throws MalformedInputException, UnmappableCharacterException {
        return getAsQueryString(key, true);
    }

    /**
     * Concatenates all values for the given key to a list of key/value pairs
     * suitable for use in a URL query string.
     *
     * @param key the parameter name
     * @param percentEncode whether key should be percent encoded before being
     *        used with the map
     * @return the query string
     * @throws MalformedInputException if input is malformed
     * @throws UnmappableCharacterException if characters are unmappable
     */
    public String getAsQueryString(String key, boolean percentEncode)
            throws MalformedInputException, UnmappableCharacterException {
        String k = percentEncode ? percentEncoder.encode(key) : key;
        Collection<String> values = multiMap.asMap().get(key);
        Iterator<String> it = values.iterator();
        StringBuilder sb = new StringBuilder();
        while (it.hasNext()) {
            String v = it.next();
            v = percentEncode ? percentEncoder.encode(v) : v;
            sb.append(k).append(EQUAL_CHAR).append(v);
            if (it.hasNext()) {
                sb.append(AMPERSAND_CHAR);
            }
        }
        return sb.toString();
    }

    public void addPercentEncodedBody(String body) {
        if (body == null) {
            return;
        }
        // watch out for "plus" encoding, replace it with a space character
        String s = body.replace('+', ' ');
        while (s != null) {
            Pair<String, Object> pairs = indexOf(AMPERSAND_CHAR, s);
            Pair<String, Object> pair = indexOf(EQUAL_CHAR, pairs.getKey());
            if (pair.getKey() != null && !pair.getKey().isEmpty()) {
                try {
                    String key = percentDecoder.decode(pair.getKey());
                    String value = pair.getValue() instanceof CharSequence ? percentDecoder.decode((CharSequence) pair.getValue()) : pair.getValue().toString();
                    addRaw(key, value);
                } catch (MalformedInputException | UnmappableCharacterException e) {
                    logger.log(Level.WARNING, "unable to decode parameter in body: " + e.getMessage(), e);
                }
            }
            s = pairs.getValue() !=null ? pairs.getValue().toString() : null;
        }
    }

    private static Pair<String, Object> indexOf(char ch, String input) {
        int i = input.indexOf(ch);
        String k = i >= 0 ? input.substring(0, i) : input;
        Object v = i >= 0 ? input.substring(i + 1) : null;
        return Pair.of(k, v);
    }
}
